/*dijkstra.c*/

/* Copyright 2004-2005 by Gianfranco Gignina <shinkuro@amiga.it>
 * All Rights Reserved.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*/

/* Implementazione sottoforma di funzione dell'algoritmo di Dijkstra per
   calcolare il percorso minimo tra due nodi di un grafo orientato;
   La funzione richiede il numero di nodi N che compongono il grafo, la
   variabile di tipo char DG [n/s] che indica se il grafo da analizzare sia o
   meno un DGrafo, la matrice dei pesi P[][] che rappresenta il grafo da
   analizzare, un intero Np rappresentante il nodo di partenza ed un intero Na
   rappresentante il nodo di arrivo (seguendo lo standard del file di tipo
   *.grf, ciascuna componente della matrice dei pesi rappresenta un nodo, dunque
   la componente P[0][1] rappresenta l'arco AB, la componente P[0][2]
   rappresenta l'arco AC e cosivvia...), infine la funzione dijkstra() richiede
   il puntatore ad un vettore di tipo intero dove verranno memorizzati tutti i
   nodi (le tappe) che comporranno il percorso minimo tra il nodo di partenza e
   il nodo di arrivo; la funzione dijkstra() ritorna il numero di nodi che
   compongono il percorso minimo comprendente il nodo di partenza e il nodo di
   arrivo. Se la funzione non riesce ad assolvere la computazione allora ritorna
   il valore NonE;
   L'algoritmo utilizza principalmente due vettori di strutture di tipo
   'elemTab': ciascuna struttura di tipo elemTab rappresenta un elemento di una
   tabella che nella teoria dell'algoritmo di Dijkstra viene costituita(vedere
   l'header file RicercaOperativaDef.h per maggiori delucidazioni riguardo tal
   tipo di struttura);
   Pi precisamente il primo vettore attuale[] rappresenta la riga della tabella
   che viene analizzata nel passo k, mentre il vettore precedente[] rappresenta
   la riga della tabella che  stata analizzata nel passo k-1;
   In principio viene considerato l'elemento della tabella costituito dal nodo
   di partenza: si effettua una copia della variabile Np nella
   variabile intera Vp, in questo modo si preserva il valore della variabile Np;
   Viene effettuata con lo stesso scopo la copia di Na nella variabile intera
   Va;
   I vettori attuale[] e precedente[] vengono in principio inizializzati con
   valori come NonE;
   Viene inizializzato l'elemento attuale[Vp] con i valori presi opportunamente
   dalla matrice dei pesi: attuale[Vp] rappresenta l'elemento di partenza della
   prima riga della tabella dell'algoritmo di Dijkstra, il quale ha in s il
   nodo di partenza;
   A questo punto inizia un ciclo di tipo while:
       Si prende in considerazione attuale[Vp] e il nodo che risiede in Vp, e si
       calcola la stella di questo nodo attraverso un ciclo for, facendo
       attenzione a non considerare  quegli elementi (e quindi anche quei nodi)
       che sono gi stati analizzati (gli elementi considerati, cio gli
       asteriscati nella tabella di Dijkstra, anche se appartenenti alla stella
       di Vp, non sono pi utili allo scopo dell'algoritmo);
       Si effettua quindi un secondo ciclo for che confronta gli elementi del
       vettore attuale[] con gli elementi del vettore precedente[], escudendo
       sempre gli elementi in precedenza considerati; Lo scopo di questo ciclo
       for  il confrontare i percorsi pi convenienti sia in attuale[]
       che in precedente[], in modo da poter scartare quegli elementi della
       tabella aventi percorsi che sono maggiori rispetto ai percorsi che gli
       stessi elementi avevano fatto nel passo precedente; Dopo questo ciclo
       il vettore attuale[] conterr degli elementi avanti i percorsi pi
       convenienti fino a quel momento;
       Viene effettuata quindi una copia del vettore attuale[] in un vettore di
       tipo elemTab chiamato Temp[] (escludendo sempre gli elementi considerati)
       sul quale verra operato un ordinamento mediante la funzione qsort()
       rispetto al percorso di ciascun elemento del vettore; Dopo l'azione di
       qsort() nella prima posizione di Temp[] si trover l'elemento del
       vettore attuale[] che possiede il percorso pi breve;
       A questo punto si effettua un ulteriore ciclo for per cercare
       l'elemento Temp[primo]  nel vettore attuale. Quando questo elemento
       verr trovato in attuale[] allora lo si 'marcher' (cio diventer un
       elemento asteriscato) in modo che rappresenti un elemento gi
       considerato; A questo punto viene effettuata la copia del vettore
       attuale[] nel vettore precedente[], dopodich il ciclo while ricomincia;
  Il ciclo while si interrompe nel momento in cui l'elemento che risiede nel
  vettore attuale[], rappresentante il nodo di arrivo, viene marcato(o
  astriscato);
  Alla fine del ciclo while il vettore attuale conterr tutti gli elementi
  aventi i nodi per giungere dal nodo di partenza al nodo di arrivo nel modo pi
  conveniente: tale vettore rappresenta adesso l'ultima riga della tabella di
  Dijkstra;
  Per poter ricavare dal vettore attuale[] il percorso minimo che va dal nodo
  di partenza al nodo di arrivo si effettua un procedimento al rovescio:
  se copia il nodo di arrivo nella prima posizione del vettore di interi V[],
  dopodich si considera l'elemento che nel vettore attuale[] rappresenta il
  nodo di arrivo: da tale struttura si prende in considerazione la variabile
  .vPrec che rappresenta il nodo precedente partendo dal nodo di arrivo, tale
  nodo precedente viene memorizzato nella seconda posizione di V[], dopodich
  si considera questo nodo precedente , si considera l'elemento che nel vettore
  attuale[] rappresenta tale nodo precedente e si effettua lo stesso
  procedimento fin quando non si giunge al nodo di partenza, a questo punto
  all'interno del vettore V[] ci sar il minimo percorso che collega il nodo
  di partenza con il nodo di arrivo, ma tale percorso  memorizzato in senso
  contrario rispetto a come dovrebbe essere attraversato. Per tale motivo viene
  incrementata una variabile di tipo intero nV, la quale rappresenta il numero
  di nodi che compongono il percorso minimo trovato;
  Attraverso la variabile nV e una variabile di tipo intero varTemp viene
  riordinato il vettore V[] che dopo tale processo conterr in ordine corretto
  i nodi da attraversare per effettuare il percorso minimo dal nodo di partenza
  al nodo di arrivo;
  La funzione dijkstra() ritorna il nomero di nodi che compongono il percorso
  minimo tra i nodi di partenza e di arrivo immessi, includendo questi ultimi;
*/

#include <stdio.h>
#include "RicercaOperativaDef.h"
#define partenza 0
#define primo 0
enum {no, si};

int dijkstra(int N, int DG, float P[][maxM], int Np, int Na, int *V)
{
/*DICHIARAZIONI*/
//------------------------------------------------------------------------------
    int i, j, Vp, Va, nV=0, varTemp;
    elemTab precedente[N], attuale[N], Temp[N];
//------------------------------------------------------------------------------

/*Controllo se il grafo  un DGrafo*/
//------------------------------------------------------------------------------
    if (DG == 'n') return NonE;//Se il grafo immesso non  un DGrafo allora la
                               //funzione si blocca e ritorna NonE;
//------------------------------------------------------------------------------

/*Preservo il nodo di partenza e il nodo di arrivo*/
//------------------------------------------------------------------------------
    Vp=Np;
    Va=Na;
//------------------------------------------------------------------------------

/*Inizializzazione dei vettori precedente[] e attuale[]*/
//------------------------------------------------------------------------------
    for (i=0; i<N; i++)
    {
        precedente[i].vPrec=attuale[i].vPrec=NonE;
        precedente[i].misPerc=attuale[i].misPerc=NonE;
        precedente[i].considerato=attuale[i].considerato=no;
    }
//------------------------------------------------------------------------------

/*Primo passo dell'algoritmo*/
//------------------------------------------------------------------------------
    attuale[Vp].vPrec=Vp;//Nel nodo precedente va il nodo di partenza
    attuale[Vp].misPerc=partenza;
    attuale[Vp].considerato=si;//L'elemento viene marcato;
//------------------------------------------------------------------------------

    while (attuale[Na].considerato==no) //fino a quando l'elemento del vettore
    {                                   //attuale[] che rappresenta il nodo di
                                        //arrivo non  stato considerato allora
                                        //inizia questo ciclo;
//------------------------------------------------------------------------------

/*Ricerca della stella del nodo all'interno dell'elemento considerato*/
//------------------------------------------------------------------------------
        for (j=0; j<N; j++)
        {

            if (P[Vp][j]==NonE) continue;// scontato per costruzione...;
            if (attuale[j].considerato==si) continue;//...idem ma per coerenza;
            attuale[j].misPerc=attuale[Vp].misPerc+P[Vp][j];//Il percorso
            /*IN QUESTA IMPLEMENTAZIONE IL PESO DI   */     //da attraversare 
            /*UN ARCO RAPPRESENTA ANCHE LA LUNGHEZZA */     //la somma tra il
            /*DI UN ARCO;                            */     //percorso gi
                                                            //attraversato e il
                                                            //peso dell'arco
                                                            //appartenente alla
                                                            //stella di Vp;

            attuale[j].vPrec=Vp;//Il nodo precedente  ovviamente Vp;
        }
//------------------------------------------------------------------------------

/*Controllo sulla convenienza dei percorsi scelti*/
//------------------------------------------------------------------------------
        for (i=0; i<N; i++)
        {
            if (attuale[i].considerato == si) continue;
            if (attuale[i].misPerc < precedente[i].misPerc){
                continue;} //Se il percorso di un elemento di attuale[] minore
                           //rispetto a quello dello stesso elemento
                           //nel passo precedente rappresentato dall'elemento
                           //che risiede nel vettore precedente allora non fare
                           //nulla...altrimenti....
            attuale[i].misPerc=precedente[i].misPerc;//...sostituisci l'elemento
            attuale[i].vPrec=precedente[i].vPrec;    //del vettore attuale[] con
        }                                            //l'elemento corrispondente
                                                     //che si trova nel vettore
                                                     //precedente[];
//------------------------------------------------------------------------------

/*Ricerca l'elemento dal percorso pi breve*/
//------------------------------------------------------------------------------
        for (i=0; i<N; i++)//inizializzo il vettore temporaneo Temp[];
        {
            Temp[i].vPrec=NonE;
            Temp[i].misPerc=NonE;
            Temp[i].considerato=no;
        }

        for (i=0; i<N; i++)//copio in Temp[] il vettore attuale[]...
        {                  //...tralasciando gli elementi che sono gi stati
            if (attuale[i].considerato==si) continue;//considerati;
            Temp[i]=attuale[i];
        }
        qsort(Temp, N, sizeof(elemTab),//Trovo in Temp[] l'elemento con il
                  (int (*)(void const *, void const *))comparePerc);//percorso
                                                                    //minore;
//------------------------------------------------------------------------------

/*Ricerca in attuale[] l'elemento trovato con qsort()*/
//------------------------------------------------------------------------------
        for (i=0; i<N; i++)
        {
            if (attuale[i].considerato == si) continue;
            if (attuale[i].vPrec != Temp[primo].vPrec) continue;
            if (attuale[i].misPerc != Temp[primo].misPerc) continue;

            if (attuale[i].misPerc == NonE) return NonE;//se attuale contiente
            /*LIMITAZIONE DI QUESTA IMPLEMENTAZIONE!*/  //un percorso pari al
                                                        //valore NonE allora
                                                        //la funzione si blocca
                                                        //e ritorna NonE!;

            attuale[i].considerato=si;//Marco l'elemento trovato perch contiene
                                      //il percorso pi conveniente;

            Vp=i;//i rappresenta in questo caso il nodo in cui si trova l'arco
                 //dal percorso(peso in questo caso) pi conveniente. Per tale
                 //motivo viene copiato sulla variabile Vp, sovrascrivendo il
                 //precendente valore (al primo passo del while Vp contiene il
                 //nodo di partenza;

            break;// Fermo il ciclo for perch ho trovato quello che cercavo;
        }
//------------------------------------------------------------------------------

/*Copia nel vettore precedente[] il vettore attuale[]*/
//------------------------------------------------------------------------------
        for (j=0; j<N; j++) {precedente[j]=attuale[j];}
    }
//------------------------------------------------------------------------------

/*Costruzione del vettore V[], PASSO-1*/
//------------------------------------------------------------------------------
    V[primo]=Va;//Nella prima posizione di V[] metto il nodo di arrivo;
    Vp=Np;//Recupero il nodo di partenza;
    nV++;//Con questo incremento la variabile nV = 1;
//------------------------------------------------------------------------------

/*Costruzione del vettore V[], PASSO-2*/
//------------------------------------------------------------------------------
    for (i=1; i<N-1; i++)//i=1 perch il primo nodo  gi presente in V[];
    {
        V[i]=attuale[Va].vPrec;//Nella seconda posizione di V[] andr il nodo
                               //indicato all'interno dell'elemento che 
                               //rappresentato da Va;

        Va=attuale[Va].vPrec;//Va diventa il nodo indicato all'interno di
                             //attuale[Va];

        nV++;//Incremento il numero di nodi che compongono il percorso minimo;
        if (Va == Vp) break;//Se in V[] sono stati messi tutti i nodi che
                            //compongono il percorso minimo allora ferma questo
    }                       //ciclo;
//------------------------------------------------------------------------------

/*Ordinamento del vettore V[]*/
//------------------------------------------------------------------------------
    j=nV;//Metto il valore di nV in j per migliorare la lettura del sorgente;

    for (i=0; i<nV/2; i++) //In pratica effettuo degli scambi tra i primi e
    {                      //gli ultimi elementi del vettore V[]: dato che uno
        varTemp=V[i];      //scambio rappresenta la considerazione di due nodi
        V[i]=V[j-1];       //allora il ciclo si conclude quando si raggiunge
        V[j-1]=varTemp;    //la met del vettore;
        j--;
    }
//------------------------------------------------------------------------------

    return nV;//Ritorna il numero di nodi che compongono il percorso minimo;
}
