/*Algoritmo di Kruscal*/

/* Copyright 2004-2005 by Gianfranco Gignina <shinkuro@amiga.it>
 * All Rights Reserved.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*/

/*L'algoritmo qui implementato utilizza la matrice dei pesi P[][] e il vettore
  degli archi E[] per cercare il minimo albero ricoprente del grafo 
  rappresentato da queste due strutture;
  kruscal() produce un vettore di strutture 'arco' che compongono il minimo
  albero ricoprente e ritorna il numero di questi archi;
  Dopo aver ottenuto il vettore di strutture 'arco' E[] dalla funzione 
  convertMtrxToVE(), l'algoritmo di Kruscal riordina il vettore E[] in ordine 
  crescente rispetto al peso degli archi in esso contenuti, -a questo proposito 
  utilizza una particolare versione dell'algoritmo bubblesort-
  Dopo aver ottenuto un vettore ordinato, questa funzione inizializza due 
  vettori di interi nodiI[] e nodiJ[] al valore 'NonE', inizializza x, k e flag 
  a zero in modo da entrare in un ciclo while che conterr un ciclo for che 
  nelle prime due iterazioni non avr alcun effetto (lo scarto di un arco che 
  forma un ciclo non  infatti necessario con due soli archi), a questo punto 
  inserisce nel vettore di archi Kr[] il primo arco di peso minore e i punti 
  incidenti ad esso rispettivamente nelle prime posizioni di nodiI[] e nodiJ[]. 
  Dopo questa istruzione controlla se la variabile x (che indica il numero di 
  archi del minimo albero ricoprente) corrisponde al numero di nodi meno 1 (cio 
  se gli archi all'interno del vettore Kr[] tocchino o meno tutti i nodi del 
  grafo).
  Il controllo effettuato per mezzo del ciclo for con il supporto della 
  variabile flag serve a verificare che un possibile arco da includere nel 
  minimo albero ricoprente non formi ciclo, a tal proposito sono utili i 
  due vettori nodiI[] e nodiJ[] i quali mantengono in memoria tutti i nodi che 
  fanno gi parte dell'albero in costruzione: se un arco in questione tocca 
  entrambi i nodi indicati in nodiI[] e nodiJ[] allora former un ciclo con 
  gli archi gi presenti nell'albero in costruzione. Se avviene quest'ultima 
  situazione allora flag assumer il valore 2 e l'arco analizzato verr scartato 
  e si passer ad analizzare il successivo arco presente nel vettore E[]. 
  L'algoritmo si conclude quando tutti gli archi presenti in Kr[] toccano tutti 
  i nodi del grafo di partenza;                 
*/


#include <stdio.h>
#include "RicercaOperativaDef.h"

int kruscal(int N, char DG, float P[][maxM], arco *E)
{
/* Dichiarazioni ed inizializzazioni */    
//------------------------------------------------------------------------------
    int i, j, x, y, k, flag, nArchi = 0;
    int nodiI[maxM]; //rappresenterebbero le righe della matrice dei pesi;
    int nodiJ[maxM]; //rappresenterebbero le colonne della matrice dei pesi;
    
    arco Kr[maxM*maxM]; //vettore temporaneo in cui verranno depositati gli 
                        //archi che formeranno il minimo albero ricoprente
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
    nArchi= convertMtrxToVE(N, DG, P, E);//converto la matrice dei pesi in un 
                                         //vettore di strutture 'arco';
//------------------------------------------------------------------------------    

//------------------------------------------------------------------------------

/* Ordinamento del vettore E[] in ordine crescente secondo il peso degli archi 
   in esso contenuti*/
//------------------------------------------------------------------------------    
    bubsortE(E, nArchi); // Ordinamento archi con bubblesort
//------------------------------------------------------------------------------

/* Inizializzazioni*/    
//------------------------------------------------------------------------------    
    x=0;
    k=0;
    flag=0; // Il valore di flag indica all'algoritmo se un arco analizzato 
            // forma un ciclo con gli archi che formano l'albero in costruzione 
            // in Kr[]
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------    
    for (i=0; i<N;i++)  // Questa inizializzazione viene effettuata come 
    {                   // prevenzione verso un possibile valore uguale
        nodiI[i]=NonE;  // a qualche nodo del grafo: se accadesse una cosa
        nodiJ[i]=NonE;  // del genere la variabile flag potrebbe assumere
    }                   // valori sbagliati invalidando cos l'intero algoritmo;
//------------------------------------------------------------------------------    

/* Controllo, inserimento e costruzione del minimo albero ricoprente*/
//------------------------------------------------------------------------------    
    while (k<nArchi)
    {
        for(y=0; y<x; y++)                 // Analisi all'arco E[k]: 
        {                                 
            if (nodiI[y]==E[k].v1) flag++; // se i nodi che incide appartengono 
            if (nodiJ[y]==E[k].v2) flag++; // all'albero in costruzione allora
        }    
        
        if (flag==2) { k++; flag=0; continue; }// l'arco E[k] verr scartato e 
                                               // ci si sposter di una 
                                               // posizione nel vettore E[] 
                                               // degli archi del grafo...
                                                           
        Kr[x]=E[k];// ..altrimenti E[k] verr incluso nell'albero in costruzione
        nodiI[x]=E[k].v1;
        nodiJ[x]=E[k].v2;
        flag=0;
        k++; // spostamento di una posizione nel vettore E[] degli archi del grf 
        x++; // incremento del num di archi componenti l'albero in costruzione;
                      
        if (x == N-1) break; // Se gli archi dell'albero in costruzione toccano    
    }                        // tutti i nodi del grafo di partenza allora si 
                             // giunti al minimo albero ricoprente;
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------  
    for (i=0; i<x; i++) {E[i]=Kr[i];} // copio il minimo albero ricoprente sul
                                      // vettore E[]...
       
    return x; // ...e ritorno il numero di archi che compongono 
              // il minimo albero ricoprente;
//------------------------------------------------------------------------------    
    
}    
