/*prim.c (Implementazione dell'algoritmo di Prim)*/

/* Copyright 2004-2005 by Gianfranco Gignina <shinkuro@amiga.it>
 * All Rights Reserved.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*/

/* Questo algoritmo, implementato sottoforma di una funzione, calcola il minimo
   albero ricoprente di un grafo. L'algoritmo si serve principalmente di una 
   matrice dei pesi P[][] rappresentante il grafo da analizzare. I dati 
   all'interno della matrive verranno elaborati in modo da produrre un vettore
   di strutture di tipo 'arco' che rappresenter tutti gli archi appartenenti
   al minimo albero ricoprente del grafo in esame. 
   Si tenga presente che la matrice dei pesi P[][] passata per indirizzo verr 
   modificata e dunque dopo lo svolgimento di questa funzione e non potr pi 
   essere considerata in un programma chiamante a meno di aver fatto un 
   opportuno backup di questa matrice prima di chiamare la funzione prim();
   La funzione prim() richiede in entrata anche il numero di nodi e la variabile
   char DG [n/s] che indica se il grafo in esame rappresenti o meno un DGrafo.
   L'algoritmo si svolge nel modo seguente:
       vengono utilizzati due vettori di interi VG[] e ST[] che rispettivamente 
       conterranno l'insieme dei nodi non ancora appartenenti al minimo albero
       ricoprente e l'insieme dei nodi facenti parte del minimo albero
       ricoprente in costruzione.
       Viene scelto un nodo arbitrario dall'insieme VG[] e viene inserito
       nell'insieme ST[]; Tale primo nodo  da considerarsi la radice del minimo
       albero ricoprente che si andr a costruire.
       Per tener conto del numero di nodi gi presenti nell'insieme ST[] viene 
       incrementata la variabile di tipo intero NumNodiST.
       A questo punto l'algoritmo effettua, per ogni nuovo nodo inserito
       all'interno dell'insieme ST[], una ricerca all'interno della matrice dei 
       pesi P[][]: si considerano tutti gli archi che compongono la stella del
       nodo in analisi che si trova all'interno dell'insieme ST[], escludendo da 
       quest'analisi gli archi che sono tangenti con altri nodi all'interno
       dell'insieme ST[]. Ciascun arco "sicuro"(cio un arco che connette 
       un nodo interno a ST[] ed un nodo esterno ad esso) verr memorizzato
       all'interno di una struttura 'arco' in modo da non perdere le 
       informazioni riguardanti i nodi che lo compongono. 
       L'intera stella di archi di un nodo interno a ST[] former cos un 
       vettore di strutture 'arco' chiamato Temp[]. Quest'ultimo vettore verr 
       ordinato mediante l'algoritmo qsort() della libreria standard stdlib.h, 
       prendendo in considerazione il peso di ciascun arco. Tale ordinamento 
       permette di avere nella prima posizione del vettore Temp[] 
       l'arco dal peso minore che incide con il nodo in esame, questo arco 
       verr memorizzato nel vettore di strutture 'arco' 
       TempGlob[]. Il vettore TempGlog[] conterr di ciascun nodo interno
       all'insieme ST[] l'arco incidente dal peso minore.
       Viene effettuato un secondo ordinamento, questa volta di TempGlob[], in
       tal modo si avr nella prima posizione di questo vettore l'arco dal peso
       minore che connette l'insieme ST[] con l'insieme VG[]. Da questo arco 
       si ricaver il nodo esterno ad ST[] che attraverso tale procedimento
       verr ora incluso nell'insieme ST[].
       L'algoritmo infine memorizzer l'arco trovato nel vettore 'arco' chiamato
       archiST[] (passato per indirizzo alla funzione). 
       Per non analizzare gli stessi archi e per evitare di considerare archi
       formati da nodi interni ad ST[] viene effettuato un ultimo ciclo for
       che rimuove dalla matrice dei pesi gli archi da non considerare.
       L'algoritmo incrementer quindi la variabile NumNodiST.
       L'algoritmo itera questi passi e si conclude nel momento in cui la 
       variabile NumNodiST diventa uguale al numero di nodi N.
       La funzione prim() ritorna un valore intero corrispondente al numero di
       archi che compone il minimo albero ricoprente.
*/
#define arbitrario 0
#define primo 0
#include <stdio.h>
#include <stdlib.h>
#include "RicercaOperativaDef.h"

int prim(int N, char DG, float P[][maxM], arco *archiST)
{
/*DICHIARAZIONI*/
//------------------------------------------------------------------------------    
    int i, j, t=0, NumNodiST=0, nArchi=0, stella=0;
    int VG[maxM], ST[maxM];

    arco Temp[N-1], TempGlob[N-1];
//------------------------------------------------------------------------------

/*Inizializzazione*/    
//------------------------------------------------------------------------------    
    for (i=0; i<N; i++) {VG[i]=i;}//Inizializzazione dell'insieme VG[]: ciascuna
                                  //posizione di questo vettore conterr un
                                  //nodo che compone il grafo in esame. Ciascun
                                  //nodo , per costruzione, un indice della
                                  //matrice dei pesi;
//------------------------------------------------------------------------------    

/*Primo passo dell'algoritmo*/    
//------------------------------------------------------------------------------    
    ST[primo]=VG[arbitrario];//Viene inserito in ST[] un nodo arbitrario  
    NumNodiST++;             //preso da VG[] e viene incrementata la variabile
                             //NumNodiST; 
//------------------------------------------------------------------------------

/*Ricerca arco dal peso minimo che connette ST[] a VG[]*/
//------------------------------------------------------------------------------    
    while (NumNodiST != N)
    {
        for (i=0; i<NumNodiST; i++) //per ogni nodo presente in ST[]
        {
            for (j=0; j<N; j++)
            {
                if (P[ST[i]][j]==NonE) continue;//Scarta le non relazioni e
                                                //gli archi non validi; 
                
                Temp[stella].v1=ST[i];//Memorizzazione del primo nodo che
                Temp[stella].v2=j;//compone l'arco:  il nodo interno ad ST[],    
                Temp[stella].peso=P[ST[i]][j];//mentre l'altro nodo in .v2 sar
                                              // il nodo in VG[]; 
                stella++;//Incremento la variabile stella;
            }
            qsort(Temp, stella, sizeof(arco),//Trovo in Temp[] l'arco dal peso 
                  (int (*)(void const *, void const *))compareArcP);//minore;
                  
            TempGlob[i]=Temp[primo];//Memorizzo l'arco dal peso minore del nodo
                                    //considerato in TempGlob[]
            stella=0;//reinizializzo la variabile stella per un successivo uso;
            t++;//la variabile ti indica quanti archi ci sono in TempGlob[];      
        }
//------------------------------------------------------------------------------

/*Ricerca e salvataggio dell'arco dal peso minimo che connette ST[] a VG[]*/
//------------------------------------------------------------------------------        
        qsort(TempGlob,t,sizeof(arco),//Trovo in TempGlob[] l'arco dal peso
              (int (*)(void const *, void const *))compareArcP);//minore;
        
        t=0;//reinizializzo la variabile t per un successivo uso;
        
        ST[NumNodiST]=TempGlob[primo].v2;//L'arco minore  stato trovato, dunque
        //viene memorizzato in ST[] il nodo in VG[] che incide con quest'arco; 
        archiST[nArchi]=TempGlob[primo];//memorizzo l'arco trovato;
        nArchi++;//incremento il numero di archi trovati
//------------------------------------------------------------------------------

/*Rimozione da P[][] dell'arco trovato e da VG[] gli archi non sicuri*/
//------------------------------------------------------------------------------       
        for (j=0; j<NumNodiST; j++)
        {
            P[ST[j]][ST[NumNodiST]]=NonE;
            
            if (DG=='n') P[ST[NumNodiST]][ST[j]]=NonE;//Se il grafo non  un    
        }//DGrafo allora P[][]  simmetrica e si deve rimuovere l'arco
         //simmetrico all'arco trovato;
//------------------------------------------------------------------------------
        
        NumNodiST++; //Incremento;        
    } 
    
    return nArchi; //Valore di ritorno;                      
}
