/*regine.c (Implementazione in C dell'algoritmo delle n Regine)*/

/* Copyright 2004-2005 by Gianfranco Gignina <shinkuro@amiga.it>
 * All Rights Reserved.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*/

/*Questa funzione implementa l'algoritmo per la disposizione corretta di n
  regine su di una scacchiera mediante la tecnica del backtracking; La funzione
  accetta in ingresso una variabile intera NumQueen indicante il numero di
  regine da disporre e una matrice di interi ChessBoard[][] che rappresenta la
  scacchiera; La funzione ritorna 0 se ha ultimato la disposizione altrimenti
  ritorna 1;
  Dopo aver inizializzato alcune variabili ed aver eseguito il controllo sul
  numero di regine da disporre la funzione regine() chiama la funzione
  clearboard() per svuotare la scacchiera (riempie la matrice di zeri), a questo
  punto incomincia un ciclo while che contiene inizialmente una parte di codice
  che si si occupa del backtracking:
      se la variabile row ha superato il numero di righe che
      compongono la matrice (cio ci troviamo ad aver controllato tutte le
      possibili posizioni di una colonna che risultano quindi non sicure per
      la regina inserita) allora parte il processo di backtracking:
      si decrementa di 1 la variabile q che rappresenta il numero di regine
      sin'ora disposte correttamente e si cerca nella colonna precedente colPrec
      alla colonna in cui ci si trova (colPrec=column-1;) la regina di tale
      colonna precedente;
      a questo punto si rimuove tale regina trovata e ci si sposta, rispetto
      alla regina rimossa, di una riga (row++); se dopo quest'azione ci si
      accorge di aver comunque superato l'ultima riga della matrice allora si
      riprende questo stesso processo di backtracking mediante l'istruzione
      'continue' che permette di rieseguire il ciclo while fin'ora percorso;

  Viene dunque inserita una regina nella scacchiera in posizione
  ChessBoard[row][column], a questo punto iniziano due controlli principali che
  permettono di capire se nella posizione in cui la regina  stata inserita
  potrebbe essere 'mangiata' da un'altra regina:

      1) Controllo sulla riga:
         Tale controllo non fa altro che sommare all'interno della variabile
         intera varTemp tutti i numeri presenti nella riga in cui la regina 
         stata inserita; Se varTemp risulta contenere pi di una regina allora
         la regina inserita viene tolta, ci si sposta di una riga (row++) e si
         ricomincia il ciclo while;

      2) Controllo sulle diagonali:
         Questo  un controllo un p pi complicato difatti si inizia un ciclo
         for attraverso l'indice i, all'interno di questo ciclo si utilizzano le
         variabili temporanee x, y, z e w che rappresenteranno:
             [x][y]= Il controllo sulla prima diagonale '\' che parte dal centro
                     della diagonale stessa (corrispondente alla posizione
                     della regina inserita) e la analizza sino al basso;

             [z][w]= Il controllo sulla prima diagonale '\' che parte dal centro
                     della diagonale stessa e la analizza fino in cima;

             [z][y]= Il controllo sulla seconda diagonale '/' che parte dal
                     centro della diagonale stessa e la analizza sino al basso;

             [x][w]= Il controllo sulla seconda diagonale '/' che parte dal
                     centro della diagonale stessa e la analizza sino all'alto;

         I controlli seguono tutti la stessa filosofia: l'algoritmo tiene conto
         della dimensione della scacchiera ed inserisce nella variabile varTemp2
         la somma di tutti i valori presenti sulla diagonale principale eccetto
         la regina considerata e inserisce nella variabile varTemp3 la somma di
         tutti i valori presenti sulla diagonale secondaria eccetto la regina
         considerata; Se varTemp2 o varTemp3 risultassero contenere almeno una
         regina allora significherebbe che la posizione in cui si  disposta
         la regina considerata la farebbe 'mangiare' da una precedente regina
         che si trova o sulla diagonale principale o su quella secondaria
         (sempre rispetto alla regina considerata), per tale motivo si rimuove
         la regina considerata dalla sua posizione, ci si sposta di una riga
         rispetto alla posizione in cui ci si trova (row++) e si ripercorre il
         ciclo while;

  Alla fine di queste sezioni di codice il ciclo while incrementa la variabile
  q indicando cos che una regina  stata disposta correttamente sulla
  scacchiera(in pratica la regina inserita risulta in una posizione sicura),
  ci si sposta dunque di una colonna (coulumn++) e si riesegue il ciclo while;

  Il ciclo while si conclude nel momento in cui tutte le regine da disporre sono
  state posizionate in modo corretto.
*/

#include <stdio.h>
#include "Regine.h"
#include <stdlib.h>

int regine(int NumQueen, int ChessBoard[][maxM])
{
    int q, x, y, z,  w, tempVar, tempVar2=0, tempVar3=0;
    int i, j;
    int row, column, colPrec;

    q=row=column=colPrec=tempVar=0;
/*Controllo sul numero di regine da disporre*/
//------------------------------------------------------------------------------
    if (NumQueen <= 3 || NumQueen > 30) return 1;
    if (NumQueen > 21)
        printf("Le regine da disporre sono molte\nAttendi prego...");
//------------------------------------------------------------------------------

/*Chiamata alla funzione che svuota la scacchiera*/
//------------------------------------------------------------------------------
    if (clearboard(NumQueen, ChessBoard)== -1) return -1;
//------------------------------------------------------------------------------


    while (q < NumQueen)
    {
        if (SetSignal(0,SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C) return -1;

/*BACKTRACKING*/
//------------------------------------------------------------------------------
        if (row == NumQueen)
        {
            q--;
            colPrec=column-1;

            for (i=0; i<NumQueen; i++)
            {
                if (ChessBoard[i][colPrec]==queen)
                {
                    row=i;
                    column=colPrec;//superfluo ma migliora la lettura;
                    ChessBoard[i][colPrec]=vuota;
                    break;
                }
            }
            row++;
        }
        if (row == NumQueen) continue;
//------------------------------------------------------------------------------

/*Provo ad inserire una regina nella scacchiera*/
//------------------------------------------------------------------------------
        ChessBoard[row][column]=queen;
//------------------------------------------------------------------------------

/*Controllo riga*/
//------------------------------------------------------------------------------
        for (j=0; j<NumQueen; j++) {tempVar=tempVar+ChessBoard[row][j];}

        if (tempVar > queen)
        {
            ChessBoard[row][column]=vuota;
            row++;
            tempVar=0;//la variabile temporanea viene svuotata;
            continue;
        }
        tempVar=0;//la variabile temporanea viene svuotata;
//------------------------------------------------------------------------------

/*Controllo Diagonali*/
//------------------------------------------------------------------------------
        for (i=1; i<NumQueen; i++)
        {
            x=row+i;
            y=column+i;
            z=row-i;
            w=column-i;

            if (x<NumQueen && y<NumQueen) tempVar2=tempVar2+ChessBoard[x][y];
            if (z>=vuota && w>=vuota) tempVar2=tempVar2+ChessBoard[z][w];

            if (z>=vuota && y<NumQueen) tempVar3=tempVar3+ChessBoard[z][y];
            if (x<NumQueen && w>=vuota) tempVar3=tempVar3+ChessBoard[x][w];
        }

        if (tempVar2>=queen || tempVar3>=queen)
        {
            ChessBoard[row][column]=vuota;
            row++;
            tempVar2=tempVar3=0;//le variabili temporanea vengono svuotate;
            continue;
        }
        tempVar2=tempVar3=0;//le variabili temporanea vengono svuotate;
//------------------------------------------------------------------------------

/*Posizione della regina valida*/
//------------------------------------------------------------------------------
        q++;
        column++;
        row=0;//Questo fa si che per ogni nuova colonna in cui si cerca di
    }         //inserire una regina si riparta dalla sua prima posizione;
//------------------------------------------------------------------------------

    return 0;

}

